// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

package frc.robot;

import frc.robot.constants.ArmConstants;
import frc.robot.constants.ElevatorConstants;
import frc.robot.constants.OIConstants;
import frc.robot.subsystems.Arm;
import frc.robot.subsystems.ClimberPivot;
import frc.robot.subsystems.ClimberRollers;
import frc.robot.subsystems.Drivetrain;
import frc.robot.subsystems.Elevator;
import frc.robot.subsystems.Indexer;
import frc.robot.subsystems.Manipulator;
import edu.wpi.first.wpilibj.shuffleboard.BuiltInWidgets;
import edu.wpi.first.wpilibj.shuffleboard.Shuffleboard;
import edu.wpi.first.wpilibj.shuffleboard.ShuffleboardTab;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.PrintCommand;
import edu.wpi.first.wpilibj2.command.button.CommandXboxController;

public class RobotContainer {
	private Arm arm;

	private ClimberPivot climberPivot;

	private ClimberRollers climberRollers;

	private Drivetrain drivetrain;

	private Elevator elevator;

	private Indexer indexer;

	private Manipulator manipulator;

	private CommandXboxController driver;
	private CommandXboxController operator;

	public RobotContainer() {
		arm = new Arm();

		climberPivot = new ClimberPivot();

		climberRollers = new ClimberRollers();

		drivetrain = new Drivetrain();

		elevator = new Elevator();

		indexer = new Indexer();

		manipulator = new Manipulator();

		driver = new CommandXboxController(OIConstants.kDriverControllerPort);
		operator = new CommandXboxController(OIConstants.kOperatorControllerPort);

		configureButtonBindings();

		configureShuffleboard();
	}

	private void configureButtonBindings() {
		arm.setDefaultCommand(
			arm.goToSetpoint(0, 1)
		);

		climberPivot.setDefaultCommand(
			climberPivot.goToAngle(0, 1)
		);

		climberRollers.setDefaultCommand(
			climberRollers.runRoller(0)
		);

		drivetrain.setDefaultCommand(
			drivetrain.drive(
				driver::getLeftY,
				driver::getLeftX,
				driver::getRightX,
				() -> true
			)
		);

		elevator.setDefaultCommand(
			elevator.runElevator(operator::getLeftY)
		);

		indexer.setDefaultCommand(
			indexer.runIndexer(0)
		);

		manipulator.setDefaultCommand(
			manipulator.runManipulator(0)
		);

		//Driver inputs
		driver.start().whileTrue(
			drivetrain.setXCommand()
		);

		driver.rightTrigger().whileTrue(
			manipulator.runManipulator(1)
		);

		//Operator inputs
		operator.povUp().onTrue(
			moveManipulator(
				ElevatorConstants.kElevatorL4Position, 
				ArmConstants.kArmL4Position
			)
		);

		operator.povRight().onTrue(
			moveManipulator(
				ElevatorConstants.kElevatorL3Position, 
				ArmConstants.kArmL3Position
			)
		);

		operator.povLeft().onTrue(
			moveManipulator(
				ElevatorConstants.kElevatorL2Position, 
				ArmConstants.kArmL2Position
			)
		);

		operator.povDown().onTrue(
			moveManipulator(
				ElevatorConstants.kElevatorL1Position, 
				ArmConstants.kArmL1Position
			)
		);

		operator.a().onTrue(
			coralIntakeRoutine()
		);

		operator.x().onTrue(
			algaeIntakeRoutine(true)
		);

		operator.b().onTrue(
			algaeIntakeRoutine(false)
		);
	}

	  //creates tabs and transforms them on the shuffleboard//
  private void configureShuffleboard() {
    ShuffleboardTab sensorTab = Shuffleboard.getTab(OIConstants.kSensorsTab);

    sensorTab.addDouble("ElevatorPosition", elevator::getEncoderPosition)
		.withSize(2, 1)
		.withPosition(0, 0)
		.withWidget(BuiltInWidgets.kTextView);

	sensorTab.addDouble("ArmPosition", arm::getEncoderPosition)
		.withSize(2, 1)
		.withPosition(2, 0)
		.withWidget(BuiltInWidgets.kTextView);
  }

	public Command getAutonomousCommand() {
		return new PrintCommand("NO AUTO DEFINED");
	}

	//teleop routines
	private Command coralIntakeRoutine() {
		return moveManipulator(
			ElevatorConstants.kElevatorCoralIntakePosition, 
			ArmConstants.kArmCoralIntakePosition
		)
			.andThen(manipulator.runUntilCollected(1, true));
	}

	private Command algaeIntakeRoutine(boolean l2) {
		return moveManipulator(
			l2 ? ElevatorConstants.kElevatorL2AlgaePosition : ElevatorConstants.kElevatorL3AlgaePosition, 
			l2 ? ArmConstants.kArmL2AlgaePosition : ArmConstants.kArmL3AlgaePosition
		)
			.andThen(manipulator.runUntilCollected(1, false));
	}

	private Command moveManipulator(double elevatorPosition, double armPosition) {
		if ((elevator.getEncoderPosition() > ElevatorConstants.kElevatorBracePosition && elevatorPosition > ElevatorConstants.kElevatorBracePosition) || (arm.getEncoderPosition() > ArmConstants.kArmSafeStowPosition && armPosition > ArmConstants.kArmSafeStowPosition)) { //if the elevator and target are above the brace, or the arm and target is in front of the brace, move together
			return arm.goToSetpoint(armPosition, 2)
				.alongWith(elevator.goToSetpoint(elevatorPosition, 2));
		} else if (armPosition < ArmConstants.kArmSafeStowPosition) { //if the target is behind the brace, move the elevator first
			if (arm.getEncoderPosition() < ArmConstants.kArmSafeStowPosition) { //if the arm is behind the brace at start, move the arm to a safe position, then elevator, then arm
				return arm.goToSetpoint(ArmConstants.kArmSafeStowPosition, 2)
					.andThen(elevator.goToSetpoint(elevatorPosition, 2))
						.andThen(arm.goToSetpoint(armPosition, 2));
			} else { //if the arm is in front of the brace, move elevator, then arm
				return elevator.goToSetpoint(elevatorPosition, 2)
					.andThen(arm.goToSetpoint(armPosition, 2));
			}
		} else if (arm.getEncoderPosition() < ArmConstants.kArmSafeStowPosition) { //if the arm is behind the brace, move to target, then move the elevator
			return arm.goToSetpoint(armPosition, 2)
				.andThen(elevator.goToSetpoint(elevatorPosition, 2));
		} else { //catch all that will always safely move the arm in a safe way
			return arm.goToSetpoint(ArmConstants.kArmSafeStowPosition, 2)
				.andThen(elevator.goToSetpoint(elevatorPosition, 2))
					.andThen(arm.goToSetpoint(armPosition, 2));
		}
	}
}
