package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.revrobotics.RelativeEncoder;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkBase.PersistMode;
import com.revrobotics.spark.SparkBase.ResetMode;
import com.revrobotics.spark.SparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ElevatorFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.wpilibj.DigitalInput;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ElevatorConstants;

public class Elevator extends SubsystemBase {
    protected SparkMax elevatorMotor1;
    protected SparkMax elevatorMotor2;

    protected RelativeEncoder encoder;

    private DigitalInput topLimitSwitch;
    private DigitalInput bottomLimitSwitch;

    private PIDController positionController;
    private PIDController velocityController;

    private ElevatorFeedforward feedForward;
    
    public Elevator() {
        elevatorMotor1 = new SparkMax(
            ElevatorConstants.kElevatorMotor1ID, 
            MotorType.kBrushless
        );
        
        elevatorMotor2 = new SparkMax(
            ElevatorConstants.kElevatorMotor2ID,
            MotorType.kBrushless
        );

        elevatorMotor1.configure(
            ElevatorConstants.motorConfig, 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        elevatorMotor2.configure(
            ElevatorConstants.motorConfig.follow(ElevatorConstants.kElevatorMotor1ID), 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        encoder = elevatorMotor1.getEncoder();

        topLimitSwitch = new DigitalInput(
            ElevatorConstants.kTopLimitSwitchID
        );

        bottomLimitSwitch = new DigitalInput(
            ElevatorConstants.kBottomLimitSwitchID
        );

        positionController = new PIDController(
            ElevatorConstants.kPositionControllerP, 
            ElevatorConstants.kPositionControllerI, 
            ElevatorConstants.kPositionControllerD
        );

        velocityController = new PIDController(
            ElevatorConstants.kVelocityControllerP, 
            ElevatorConstants.kVelocityControllerI, 
            ElevatorConstants.kVelocityControllerD
        );

        feedForward = new ElevatorFeedforward(
            ElevatorConstants.kFeedForwardS, 
            ElevatorConstants.kFeedForwardG, 
            ElevatorConstants.kFeedForwardV
        );
    }

    //manual command that keeps ouput speed consistent no matter the direction
    public Command runElevator(DoubleSupplier speed) {
        return run(() -> {
            double realSpeedTarget = speed.getAsDouble() * ElevatorConstants.kElevatorMaxVelocity;

            double voltsOut = velocityController.calculate(
                encoder.getVelocity(), 
                realSpeedTarget
            ) + feedForward.calculate(realSpeedTarget);

            elevatorMotor1.setVoltage(voltsOut);
        }).until(() -> topLimitSwitch.get() || bottomLimitSwitch.get());
    }

    
    //go to setpoint command
    public Command goToSetpoint(double setpoint, double timeout) {
        return run(() -> {
            double voltsOut = positionController.calculate(
                encoder.getPosition(),
                setpoint
            ) + feedForward.calculate(0);

            elevatorMotor1.setVoltage(voltsOut);
        }).until(
            () -> positionController.atSetpoint() || topLimitSwitch.get() || bottomLimitSwitch.get()
        ).withTimeout(timeout);
    }

    public double getEncoderPosition() {
        return encoder.getPosition();
    }
}
